"use strict";

const fs = require('fs');

const path = require('path');

const {
  Readable
} = require('stream');

const {
  removeKeysFromObject
} = require('../utils/object-manipulation');

const pluginName = 'stats-plugin';
const statsSchema = {
  all: true
};

function* convertObjectToStringGen(obj) {
  if (typeof obj === 'string' || typeof obj === 'number' || typeof obj === 'boolean' || obj === null) {
    yield JSON.stringify(obj);
  } else if (Array.isArray(obj)) {
    yield '[';
    let isFirst = true;

    for (let item of obj) {
      if (item === undefined) {
        item = null;
      }

      yield `${isFirst ? '' : ','}`;
      yield* convertObjectToStringGen(item);
      isFirst = false;
    }

    yield ']';
  } else {
    yield '{';
    let isFirst = true;
    const entries = Object.entries(obj);

    for (const [itemKey, itemValue] of entries) {
      if (itemValue === undefined) {
        continue;
      }

      yield `${isFirst ? '' : ','}${JSON.stringify(itemKey)}: `;
      yield* convertObjectToStringGen(itemValue);
      isFirst = false;
    }

    yield '}';
  }
}

class StatsPlugin {
  constructor({
    statsOptions = {},
    statsOutputExcludeKeys = [],
    statsFileName = 'bundle-report-integrity.json'
  }) {
    this.excludeKeysInStat = statsOutputExcludeKeys;
    this.statsOptions = Object.assign({}, statsSchema, statsOptions);
    this.statsFileName = statsFileName;
  }

  apply(compiler) {
    compiler.hooks.done.tapAsync(pluginName, (stats, callback) => {
      const statsJson = removeKeysFromObject(stats.toJson(this.statsOptions), this.excludeKeysInStat);
      this.emitStats(statsJson).on('end', () => {
        callback();
      }).on('error', e => {
        callback(e);
      });
    });
  }

  writeStatsFileInAStream(statsObj) {
    const {
      outputPath
    } = statsObj;
    const ouputFileName = path.join(outputPath, this.statsFileName);
    return this.createReadStream(statsObj).pipe(fs.createWriteStream(ouputFileName));
  }

  createReadStream(statsObj) {
    const excludeKeys = this.excludeKeysInStat;
    return new Readable({
      read() {
        let isDone = false;
        const objToStringGen = convertObjectToStringGen(statsObj, excludeKeys);

        while (!isDone) {
          const {
            done,
            value
          } = objToStringGen.next();

          if (done) {
            isDone = true;
            this.push(null);
          } else {
            this.push(value);
          }
        }
      }

    });
  }

  emitStats(statsJson) {
    return this.writeStatsFileInAStream(statsJson);
  }

}

module.exports = StatsPlugin;